# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.


import re
from urllib.parse import urlparse

from django import forms
from django.contrib import messages
from django.core.exceptions import ValidationError
from django.shortcuts import redirect
from django.urls import reverse
from django.views.generic import FormView

from .collection import Collection
from .document import Document
from .models import WorkerStats
from .url import sanitize_url, validate_url
from .utils import human_datetime, plural
from .views import AdminView


def _add_unique_patterns(existing_regex, new_patterns):
    """Add new patterns to existing regex, avoiding duplicates."""
    if not new_patterns:
        return existing_regex

    # Get existing patterns as a set for deduplication check
    existing_patterns = set()
    if existing_regex:
        existing_patterns = {pattern.strip() for pattern in existing_regex.split("\n") if pattern.strip()}

    # Filter new patterns to only include those not already present
    new_pattern_list = [pattern.strip() for pattern in new_patterns.split("\n") if pattern.strip()]
    unique_new_patterns = [pattern for pattern in new_pattern_list if pattern not in existing_patterns]

    if not unique_new_patterns:
        return existing_regex

    # Append only the unique new patterns to existing content
    if existing_regex:
        return existing_regex + "\n" + "\n".join(unique_new_patterns)
    else:
        return "\n".join(unique_new_patterns)


def queue_urls(urls, collection, show_on_homepage, crawl_scope):
    """Queue URLs for crawling with the specified collection and scope
    settings."""
    # Extract hostnames if scope modification is requested
    if crawl_scope != AddToQueueForm.CRAWL_SCOPE_NO_CHANGE:
        hostnames = set()
        for url in urls:
            try:
                parsed = urlparse(url)
                if parsed.hostname:
                    hostnames.add(parsed.hostname)
            except Exception:  # nosec B112
                continue

        # Add hostnames to appropriate regex field if we have any
        if hostnames:
            hostname_patterns = [f"^https?://{re.escape(hostname)}/.*" for hostname in hostnames]
            new_patterns = "\n".join(hostname_patterns)

            if crawl_scope == AddToQueueForm.CRAWL_SCOPE_UNLIMITED:
                # Add to unlimited_regex
                collection.unlimited_regex = _add_unique_patterns(collection.unlimited_regex, new_patterns)
            elif crawl_scope == AddToQueueForm.CRAWL_SCOPE_LIMITED:
                # Add to limited_regex
                collection.limited_regex = _add_unique_patterns(collection.limited_regex, new_patterns)

            collection.save()

    # Queue each URL
    for url in urls:
        Document.manual_queue(url, collection, show_on_homepage)


class AddToQueueForm(forms.Form):
    CRAWL_SCOPE_NO_CHANGE = "no_change"
    CRAWL_SCOPE_UNLIMITED = "unlimited"
    CRAWL_SCOPE_LIMITED = "limited"

    CRAWL_SCOPE_CHOICES = [
        (CRAWL_SCOPE_NO_CHANGE, "Keep collection settings unchanged"),
        (CRAWL_SCOPE_UNLIMITED, "Crawl entire websites (unlimited depth)"),
        (CRAWL_SCOPE_LIMITED, "Crawl websites with depth limit from collection settings"),
    ]

    urls = forms.CharField(
        widget=forms.Textarea(attrs={"style": "width: 100%; padding-right: 0", "rows": "3", "autofocus": True}),
        label="URLs to crawl",
    )
    collection = forms.ModelChoiceField(
        queryset=Collection.objects.all(),
        required=True,
        label="Collection",
        widget=forms.Select(attrs={"class": "form-control"}),
        empty_label=None,
    )
    crawl_scope = forms.ChoiceField(
        choices=CRAWL_SCOPE_CHOICES,
        initial=CRAWL_SCOPE_NO_CHANGE,
        widget=forms.RadioSelect,
        label="Crawling scope",
        help_text="Choose how to crawl the websites from these URLs",
    )
    show_on_homepage = forms.BooleanField(
        required=False,
        help_text="Display the initial document on the homepage",
        initial=True,
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            default_collection = Collection.objects.get(name="Default")
            self.fields["collection"].initial = default_collection.pk
        except Collection.DoesNotExist:
            first_collection = Collection.objects.first()
            self.fields["collection"].initial = first_collection.pk

    def clean_urls(self):
        errors = []
        urls = []
        for line_no, line in enumerate(self.cleaned_data["urls"].splitlines()):
            url = line.strip()
            if not url:
                continue

            try:
                url = sanitize_url(url)
                validate_url(url)
            except Exception as e:
                errors.append(f"Line {line_no + 1}: {e.args[0]}")

            urls.append(url)
        if errors:
            raise ValidationError("Invalid URL" + plural(len(errors)) + ":\n" + "\n".join(errors))
        return urls


class AddToQueueView(AdminView, FormView):
    template_name = "admin/add_to_queue.html"
    title = "Crawl a new URL"
    form_class = AddToQueueForm
    permission_required = "se.add_document"
    admin_site = None

    def __init__(self, *args, **kwargs):
        self.admin_site = kwargs.pop("admin_site")
        super().__init__(*args, **kwargs)

    def get_initial(self):
        initial = super().get_initial()

        # Pre-populate form fields from GET parameters
        if "urls" in self.request.GET:
            initial["urls"] = self.request.GET.get("urls")
        if "collection" in self.request.GET:
            try:
                collection_id = int(self.request.GET.get("collection"))
                if Collection.objects.filter(id=collection_id).exists():
                    initial["collection"] = collection_id
            except (ValueError, TypeError):
                pass
        if "show_on_homepage" in self.request.GET:
            initial["show_on_homepage"] = self.request.GET.get("show_on_homepage").lower() == "true"
        else:
            # Default value when no GET parameter is provided
            initial["show_on_homepage"] = True
        return initial

    def get_context_data(self, **kwargs):
        Collection.create_default()
        context = super().get_context_data(**kwargs)
        context.update(self.admin_site.each_context(self.request))
        form = context["form"]

        collections = Collection.objects.all()
        for collection in collections:
            if collection.recrawl_freq == Collection.RECRAWL_FREQ_CONSTANT:
                context["recrawl_every"] = human_datetime(collection.recrawl_dt_min)
            elif collection.recrawl_freq == Collection.RECRAWL_FREQ_ADAPTIVE:
                context.update(
                    {
                        "recrawl_min": human_datetime(collection.recrawl_dt_min),
                        "recrawl_max": human_datetime(collection.recrawl_dt_max),
                    }
                )
        return_url = reverse("admin:queue")
        return context | {
            "collections": collections,
            "return_url": return_url,
            "Collection": Collection,
            "form": form,
        }

    def form_valid(self, form):
        urls = form.cleaned_data["urls"]
        collection = form.cleaned_data["collection"]
        crawl_scope = form.cleaned_data["crawl_scope"]

        queue_urls(urls, collection, form.cleaned_data["show_on_homepage"], crawl_scope)

        url_count = len(urls)
        if url_count > 1:
            msg = f"{url_count} URLs were queued."
        else:
            msg = "URL was queued."
        WorkerStats.wake_up()
        messages.success(self.request, msg)
        return redirect(reverse("admin:crawl_queue"))



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

import logging

from django.db import migrations

from se.utils import build_multiline_re

logger = logging.getLogger("django.db.migrations")


# Constants for old CrawlPolicy recursion values
CRAWL_ALL = "always"
CRAWL_ON_DEPTH = "depth"
CRAWL_NEVER = "never"


def create_default_policy(apps, schema_editor):
    Collection = apps.get_model("se", "Collection")
    Collection.objects.get_or_create(
        unlimited_regex="(default)",
        defaults={
            "name": "Default",
            "unlimited_regex": "(default)",
            "unlimited_regex_pg": ".*",
            "recursion": CRAWL_ON_DEPTH,
        },
    )[0]


def update_default_policy(apps, schema_editor):
    Collection = apps.get_model("se", "Collection")
    default_collection = Collection.objects.get(unlimited_regex="(default)")
    default_recursion = default_collection.recursion
    if default_recursion == CRAWL_NEVER:
        default_collection.unlimited_regex = ""
        default_collection.unlimited_regex_pg = ""
        default_collection.limited_regex = ""
        default_collection.limited_regex_pg = ""
    elif default_recursion == CRAWL_ALL:
        default_collection.unlimited_regex = ".*"
        default_collection.unlimited_regex_pg = ".*"
        default_collection.limited_regex = ""
        default_collection.limited_regex_pg = ""
    else:
        default_collection.unlimited_regex = ""
        default_collection.unlimited_regex_pg = ""
        default_collection.limited_regex = ".*"
        default_collection.limited_regex_pg = ".*"
    default_collection.save(
        update_fields=["unlimited_regex", "unlimited_regex_pg", "limited_regex", "limited_regex_pg"]
    )


def get_title_label(collection):
    """Get title label from collection's unlimited_regex."""

    def plural(count):
        return "s" if count > 1 else ""

    if collection.unlimited_regex == "(default)":
        return "Default"

    if collection.unlimited_regex:
        url_regexs = [line.strip() for line in collection.unlimited_regex.splitlines()]
        url_regexs = [line for line in url_regexs if not line.startswith("#") and line]
        if len(url_regexs) == 1:
            return f"{url_regexs[0]}"
        elif len(url_regexs) > 1:
            others = len(url_regexs) - 1
            others = f"{others} other{plural(others)}"
            return f"{url_regexs[0]} (and {others})"
    return "<empty>"


def get_unique_name(Collection, base_name):
    """Get a unique name by adding a counter if needed."""
    name = base_name
    counter = 2

    while Collection.objects.filter(name=name).exists():
        name = f"{base_name} {counter}"
        counter += 1

    return name


def populate_collection_names(apps, schema_editor):
    """Populate the name field for existing collections using get_title_label
    logic."""
    Collection = apps.get_model("se", "Collection")

    for collection in Collection.objects.exclude(name="Default"):
        base_name = get_title_label(collection)
        unique_name = get_unique_name(Collection, base_name)
        collection.name = unique_name
        collection.save(update_fields=["name"])


def transform_collection_data(apps, schema_editor):
    """Transform collections from old CrawlPolicy format to new Collection
    format."""
    Collection = apps.get_model("se", "Collection")

    # Get default collection and its recursion value
    default_collection = Collection.objects.get(unlimited_regex="(default)")
    default_recursion = default_collection.recursion

    transformed = 0
    for collection in Collection.objects.exclude(id=default_collection.id).iterator():
        enabled = collection.enabled
        recursion = collection.recursion

        logger.debug(f"Processing collection {collection.id}: enabled={enabled}, recursion={recursion}")

        # Store original url_regex (now unlimited_regex)
        original_regex = collection.unlimited_regex

        # If disabled, clear all regex fields
        if not enabled:
            collection.unlimited_regex = ""
            collection.unlimited_regex_pg = ""
            collection.limited_regex = ""
            collection.limited_regex_pg = ""
            collection.save(
                update_fields=["unlimited_regex", "unlimited_regex_pg", "limited_regex", "limited_regex_pg"]
            )
            transformed += 1
            continue

        # Transform based on default collection's recursion and this collection's recursion
        if default_recursion == CRAWL_NEVER:
            if recursion == CRAWL_ALL:
                collection.unlimited_regex = original_regex
                collection.unlimited_regex_pg = build_multiline_re(original_regex)
                collection.limited_regex = ""
                collection.limited_regex_pg = ""
            elif recursion == CRAWL_ON_DEPTH:
                collection.unlimited_regex = ""
                collection.unlimited_regex_pg = ""
                collection.limited_regex = original_regex
                collection.limited_regex_pg = build_multiline_re(original_regex)
            elif recursion == CRAWL_NEVER:
                collection.unlimited_regex = ""
                collection.unlimited_regex_pg = ""
                collection.limited_regex = ""
                collection.limited_regex_pg = ""

        elif default_recursion == CRAWL_ON_DEPTH:
            if recursion == CRAWL_ALL:
                collection.unlimited_regex = original_regex
                collection.unlimited_regex_pg = build_multiline_re(original_regex)
                collection.limited_regex = ".*"
                collection.limited_regex_pg = ".*"
            elif recursion == CRAWL_ON_DEPTH:
                collection.unlimited_regex = ""
                collection.unlimited_regex_pg = ""
                collection.limited_regex = original_regex
                collection.limited_regex_pg = build_multiline_re(original_regex)
            elif recursion == CRAWL_NEVER:
                collection.unlimited_regex = ""
                collection.unlimited_regex_pg = ""
                collection.limited_regex = ""
                collection.limited_regex_pg = ""

        elif default_recursion == CRAWL_ALL:
            if recursion == CRAWL_ALL:
                collection.unlimited_regex = original_regex
                collection.unlimited_regex_pg = build_multiline_re(original_regex)
                collection.limited_regex = ""
                collection.limited_regex_pg = ""
            elif recursion == CRAWL_ON_DEPTH:
                collection.unlimited_regex = ""
                collection.unlimited_regex_pg = ""
                collection.limited_regex = original_regex
                collection.limited_regex_pg = build_multiline_re(original_regex)
            elif recursion == CRAWL_NEVER:
                collection.unlimited_regex = ""
                collection.unlimited_regex_pg = ""
                collection.limited_regex = ""
                collection.limited_regex_pg = ""

        collection.save(update_fields=["unlimited_regex", "unlimited_regex_pg", "limited_regex", "limited_regex_pg"])
        transformed += 1

    logger.info(f"Collection transformation completed: {transformed} collections transformed")


def populate_document_collections(apps, schema_editor):
    """Populate the collection field for existing documents using
    Collection.get_from_url()"""
    Document = apps.get_model("se", "Document")
    Collection = apps.get_model("se", "Collection")

    # Recreate the get_from_url logic using SQL operations like the production version
    def get_collection_from_url(url):
        from django.db import models

        # Get all collections except default, excluding empty regex
        queryset = Collection.objects.exclude(unlimited_regex="(default)")
        queryset = queryset.exclude(unlimited_regex_pg="")

        # Use PostgreSQL REGEXP_SUBSTR to find the best match
        return (
            queryset.annotate(
                match_len=models.functions.Length(
                    models.Func(
                        models.Value(url),
                        models.F("unlimited_regex_pg"),
                        function="REGEXP_SUBSTR",
                        output_field=models.TextField(),
                    )
                )
            )
            .filter(match_len__gt=0)
            .order_by("-match_len")
            .first()
        )

    # Process all documents
    total_docs = Document.objects.count()
    processed = 0

    logger.info(f"\nStarting document collection assignment: processing {total_docs} documents...")

    default_collection = Collection.objects.get(unlimited_regex="(default)")
    for document in Document.objects.iterator():
        collection = get_collection_from_url(document.url)
        if collection is None:
            collection = default_collection
        document.collection = collection
        document.save(update_fields=["collection"])
        processed += 1

        if processed % 1000 == 0:
            logger.info(f"Processed {processed}/{total_docs} documents...")

    logger.info(f"Document collection assignment completed: {processed} documents processed")


def populate_combined_regex_pg(apps, schema_editor):
    """Populate the combined_regex_pg field for all existing collections."""
    Collection = apps.get_model("se", "Collection")

    for collection in Collection.objects.all():
        # Build combined regex for cross-collection matching
        unlimited_regex_pg = collection.unlimited_regex_pg or ""
        limited_regex_pg = collection.limited_regex_pg or ""

        if unlimited_regex_pg and limited_regex_pg:
            collection.combined_regex_pg = f"{unlimited_regex_pg}|{limited_regex_pg}"
        elif unlimited_regex_pg:
            collection.combined_regex_pg = unlimited_regex_pg
        elif limited_regex_pg:
            collection.combined_regex_pg = limited_regex_pg
        else:
            collection.combined_regex_pg = ""

        collection.save(update_fields=["combined_regex_pg"])


class Migration(migrations.Migration):
    dependencies = [
        ("se", "0020_collection"),
    ]

    operations = [
        migrations.RunPython(create_default_policy, migrations.RunPython.noop),
        migrations.RunPython(populate_collection_names, migrations.RunPython.noop),
        migrations.RunPython(populate_document_collections, migrations.RunPython.noop),
        migrations.RunPython(transform_collection_data, migrations.RunPython.noop),
        migrations.RunPython(update_default_policy, migrations.RunPython.noop),
        migrations.RunPython(populate_combined_regex_pg, migrations.RunPython.noop),
    ]



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

from django import forms
from django.contrib import messages
from django.core.exceptions import ValidationError
from django.shortcuts import redirect
from django.views.generic import FormView

from .models import ExcludedUrl
from .views import AdminView


class ExcludedUrlForm(forms.Form):
    urls = forms.CharField(
        label="URLs",
        widget=forms.Textarea(attrs={"placeholder": "Enter URLs here, one per line...", "autofocus": True}),
        required=True,
        help_text="Enter one URL per line",
    )
    urls.widget.attrs.update({"style": "width: 50%; padding-right: 0"})
    starting_with = forms.BooleanField(label="Exclude all URLs starting with these patterns", required=False)
    comment = forms.CharField(
        label="Comment",
        widget=forms.Textarea(attrs={"placeholder": "Optional comment for these URLs..."}),
        required=False,
        help_text="Optional comment that will be applied to all imported URLs",
    )
    comment.widget.attrs.update({"style": "width: 50%; padding-right: 0; height: 80px"})

    def clean(self):
        self.cleaned_data = super().clean()
        urls_text = self.cleaned_data.get("urls", "").strip()

        if not urls_text:
            raise ValidationError("At least one URL must be provided")

        urls = []
        for line in urls_text.split("\n"):
            line = line.strip()
            if line:
                urls.append(line)

        if not urls:
            raise ValidationError("At least one URL must be provided")

        self.cleaned_data["urls_list"] = urls
        return self.cleaned_data


class ExcludedUrlsImportView(AdminView, FormView):
    template_name = "admin/excluded_urls_import.html"
    title = "Excluded URLs import"
    form_class = ExcludedUrlForm
    permission_required = "se.change_excludedurl"

    def form_valid(self, form):
        urls_list = form.cleaned_data["urls_list"]
        starting_with = form.cleaned_data.get("starting_with", False)
        comment = form.cleaned_data.get("comment", "")

        created_count = 0

        for url in urls_list:
            excluded_url, created = ExcludedUrl.objects.get_or_create(
                url=url, defaults={"starting_with": starting_with, "comment": comment}
            )
            if created:
                created_count += 1

        if created_count > 0:
            messages.success(self.request, f"{created_count} URLs added to exclusion list.")
        else:
            messages.info(self.request, "No changes were made (all URLs already existed).")

        return redirect("admin:se_excludedurl_changelist")



# Copyright 2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

from datetime import timedelta

from django.contrib.postgres.fields import ArrayField
from django.db import models

from .browser_chromium import BrowserChromium
from .browser_firefox import BrowserFirefox
from .browser_request import BrowserRequest
from .document import Document
from .domain import Domain

BROWSER_MAP = {
    Domain.BROWSE_CHROMIUM: BrowserChromium,
    Domain.BROWSE_FIREFOX: BrowserFirefox,
    Domain.BROWSE_REQUESTS: BrowserRequest,
}


class CrawlPolicyBackup(models.Model):
    RECRAWL_FREQ_NONE = "none"
    RECRAWL_FREQ_CONSTANT = "constant"
    RECRAWL_FREQ_ADAPTIVE = "adaptive"
    RECRAWL_FREQ = [
        (RECRAWL_FREQ_NONE, "Once"),
        (RECRAWL_FREQ_CONSTANT, "Constant time"),
        (RECRAWL_FREQ_ADAPTIVE, "Adaptive"),
    ]

    HASH_RAW = "raw"
    HASH_NO_NUMBERS = "no_numbers"
    HASH_MODE = [
        (HASH_RAW, "Raw content"),
        (HASH_NO_NUMBERS, "Normalize numbers"),
    ]

    RECRAWL_COND_ON_CHANGE = "change"
    RECRAWL_COND_ALWAYS = "always"
    RECRAWL_COND_MANUAL = "manual"
    RECRAWL_CONDITION = [
        (RECRAWL_COND_ON_CHANGE, "On change only"),
        (RECRAWL_COND_ALWAYS, "Always"),
        (RECRAWL_COND_MANUAL, "On change or manual trigger"),
    ]

    CRAWL_ALL = "always"
    CRAWL_ON_DEPTH = "depth"
    CRAWL_NEVER = "never"
    CRAWL_CONDITION = [
        (CRAWL_ALL, "Crawl all pages"),
        (CRAWL_ON_DEPTH, "Depending on depth"),
        (CRAWL_NEVER, "Never crawl"),
    ]

    REMOVE_NAV_FROM_INDEX = "idx"
    REMOVE_NAV_FROM_SCREENSHOT = "scr"
    REMOVE_NAV_FROM_ALL = "yes"
    REMOVE_NAV_NO = "no"
    REMOVE_NAV = [
        (REMOVE_NAV_FROM_INDEX, "From index"),
        (REMOVE_NAV_FROM_SCREENSHOT, "From index and screenshots"),
        (REMOVE_NAV_FROM_ALL, "From index, screens and HTML snaps"),
        (REMOVE_NAV_NO, "No"),
    ]

    THUMBNAIL_MODE_PREVIEW = "preview"
    THUMBNAIL_MODE_PREV_OR_SCREEN = "prevscreen"
    THUMBNAIL_MODE_SCREENSHOT = "screenshot"
    THUMBNAIL_MODE_NONE = "none"
    THUMBNAIL_MODE = (
        (THUMBNAIL_MODE_PREVIEW, "Page preview from metadata"),
        (THUMBNAIL_MODE_PREV_OR_SCREEN, "Preview from meta, screenshot as fallback"),
        (THUMBNAIL_MODE_SCREENSHOT, "Take a screenshot"),
        (THUMBNAIL_MODE_NONE, "No thumbnail"),
    )
    url_regex = models.TextField(
        help_text="URL regular expressions for this policy. (one by line, lines starting with # are ignored)",
    )
    url_regex_pg = models.TextField()
    enabled = models.BooleanField(default=True)
    recursion = models.CharField(max_length=6, choices=CRAWL_CONDITION, default=CRAWL_ALL)
    mimetype_regex = models.TextField(default=".*")
    recursion_depth = models.PositiveIntegerField(
        default=0,
        help_text="Level of external links (links that don't match the regex) to recurse into",
    )
    keep_params = models.BooleanField(
        default=True,
        verbose_name="Index URL parameters",
        help_text='When disabled, URL parameters (parameters after "?") are removed from URLs, this can be useful if some parameters are random, change sorting or filtering, ...',
    )
    hide_documents = models.BooleanField(default=False, help_text="Hide documents from search results")

    default_browse_mode = models.CharField(
        max_length=8,
        choices=Domain.BROWSE_MODE,
        default=Domain.BROWSE_DETECT,
        help_text="Python Request is faster, but can't execute Javascript and may break pages",
    )

    snapshot_html = models.BooleanField(
        default=True,
        help_text="Archive binary files, HTML content and download related assets",
        verbose_name="Archive content 🔖",
    )
    snapshot_exclude_url_re = models.TextField(
        blank=True,
        default="",
        help_text="Regex of URL to skip related assets downloading",
        verbose_name="Assets exclude URL regex",
    )
    snapshot_exclude_mime_re = models.TextField(
        blank=True,
        default="",
        help_text="Regex of mimetypes to skip related assets saving",
        verbose_name="Assets exclude mime regex",
    )
    snapshot_exclude_element_re = models.TextField(
        blank=True,
        default="",
        help_text="Regex of HTML elements to skip related assets downloading",
        verbose_name="Assets exclude HTML regex",
    )

    thumbnail_mode = models.CharField(
        default=THUMBNAIL_MODE_PREVIEW,
        help_text="Save thumbnails to display in search results",
        choices=THUMBNAIL_MODE,
        max_length=10,
    )
    take_screenshots = models.BooleanField(
        default=False,
        help_text="Store pages as screenshots",
        verbose_name="Take screenshots 📷",
    )
    screenshot_format = models.CharField(
        max_length=3,
        choices=Document.SCREENSHOT_FORMAT,
        default=Document.SCREENSHOT_JPG,
    )

    remove_nav_elements = models.CharField(
        default=REMOVE_NAV_FROM_INDEX,
        help_text="Remove navigation related elements",
        choices=REMOVE_NAV,
        max_length=4,
    )
    script = models.TextField(
        default="",
        help_text="Javascript code to execute after the page is loaded. If an object is returned, its content will be "
        "used to overwrite the document's fields",
        blank=True,
    )
    store_extern_links = models.BooleanField(default=False, help_text="Store links to non-indexed pages")

    recrawl_freq = models.CharField(
        max_length=8,
        choices=RECRAWL_FREQ,
        default=RECRAWL_FREQ_ADAPTIVE,
        verbose_name="Crawl frequency",
        help_text="Adaptive frequency will increase delay between two crawls when the page stays unchanged",
    )
    recrawl_dt_min = models.DurationField(
        blank=True,
        null=True,
        help_text="Min. time before recrawling a page",
        default=timedelta(days=1),
    )
    recrawl_dt_max = models.DurationField(
        blank=True,
        null=True,
        help_text="Max. time before recrawling a page",
        default=timedelta(days=365),
    )
    hash_mode = models.CharField(
        max_length=10,
        choices=HASH_MODE,
        default=HASH_NO_NUMBERS,
        verbose_name="Change detection",
        help_text="Content to check for modifications",
    )
    recrawl_condition = models.CharField(
        max_length=10,
        choices=RECRAWL_CONDITION,
        default=RECRAWL_COND_MANUAL,
        verbose_name="Condition",
        help_text="Specifies the conditions under which a page is reprocessed",
    )

    auth_login_url_re = models.TextField(
        null=True,
        blank=True,
        verbose_name="Login URL regex",
        help_text="A redirection to an URL matching the regex will trigger authentication",
    )
    auth_form_selector = models.TextField(
        null=True,
        blank=True,
        verbose_name="Form selector",
        help_text="CSS selector pointing to the authentication &lt;form&gt; element",
    )
    tags = ArrayField(
        models.CharField(max_length=128),
        blank=True,
        default=list,
    )
    webhooks = ArrayField(
        models.CharField(max_length=512),
        blank=True,
        default=list,
    )


class AuthFieldBackup(models.Model):
    key = models.CharField(max_length=256, verbose_name="<input> name attribute")
    value = models.CharField(max_length=256)
    crawl_policy = models.ForeignKey("CrawlPolicyBackup", on_delete=models.CASCADE)



# Copyright 2022-2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

from django.conf import settings

from ...admin import ConflictingSearchEngineFilter
from ...models import SearchEngine
from ..builtin import UpdateBuiltinModel


class Command(UpdateBuiltinModel):
    help = "Updates Search engine shortcuts."
    doc = "This updates :doc:`user/shortcuts` in the database based on their definition in the source code."

    json_file = "sosse/search_engines.json"
    model_class = SearchEngine
    lookup_field = "short_name"
    model_name = "search engines"

    def update_existing(self, db_obj, fields):
        fields.pop("shortcut")
        return super().update_existing(db_obj, fields)

    def check_conflicts(self):
        conflicts = ConflictingSearchEngineFilter.conflicts(SearchEngine.objects.all())

        if len(conflicts):
            conflicts = conflicts.values_list("shortcut", flat=True).order_by("shortcut")
            sc = ", ".join([settings.SOSSE_SEARCH_SHORTCUT_CHAR + c for c in conflicts])
            self.stderr.write(self.style.ERROR(f"WARNING: {len(conflicts)} shortcuts are in conflict: {sc}"))
            exit(1)



# Copyright 2022-2025 Laurent Defert
#
#  This file is part of Sosse.
#
# Sosse is free software: you can redistribute it and/or modify it under the terms of the GNU Affero
# General Public License as published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Sosse is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
# the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License along with Sosse.
# If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime, timedelta, timezone
from hashlib import md5
from unittest import mock

from django.test import TransactionTestCase, override_settings
from feedparser import parse

from .browser import AuthElemFailed, SkipIndexing
from .collection import Collection
from .document import Document
from .domain import Domain
from .models import ExcludedUrl, Link, WorkerStats
from .page import Page
from .test_mock import BrowserMock


class CrawlerTest(TransactionTestCase):
    def get_default_gets(self):
        return [
            mock.call("http://127.0.0.1/robots.txt", self.collection, check_status=True),
            mock.call("http://127.0.0.1/", self.collection),
            mock.call("http://127.0.0.1/favicon.ico", self.collection, check_status=True),
        ]

    def setUp(self):
        Collection.objects.all().delete()

        self.collection = Collection.objects.create(
            name="Test Collection",
            unlimited_regex="http://127.0.0.1/",
            default_browse_mode=Domain.BROWSE_REQUESTS,
            snapshot_html=False,
            thumbnail_mode=Collection.THUMBNAIL_MODE_NONE,
            take_screenshots=False,
        )
        self.fake_now = datetime(2000, 1, 1, tzinfo=timezone.utc)
        self.fake_yesterday = self.fake_now - timedelta(days=1)
        self.fake_next = datetime(2000, 1, 1, 1, tzinfo=timezone.utc)
        self.fake_next2 = datetime(2000, 1, 1, 3, tzinfo=timezone.utc)
        self.fake_next3 = datetime(2000, 1, 1, 6, tzinfo=timezone.utc)

    def tearDown(self):
        self.collection.delete()

    def _crawl(self, url="http://127.0.0.1/"):
        # Force create the worker
        WorkerStats.get_worker(0)

        Document.queue(url, self.collection, None)
        while Document.crawl(0):
            pass

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_001_hello_world(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        self._crawl()
        self.assertTrue(
            BrowserRequest.call_args_list == self.get_default_gets(),
            BrowserRequest.call_args_list,
        )

        domain = Domain.objects.get()
        self.assertEqual(domain.browse_mode, Domain.BROWSE_DETECT)
        self.assertEqual(domain.domain, "127.0.0.1")
        self.assertEqual(domain.robots_status, Domain.ROBOTS_EMPTY)

        self.assertEqual(Document.objects.count(), 1)
        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "Hello world")
        self.assertEqual(doc.crawl_recurse, 0)
        self.assertFalse(doc.hidden)

        self.assertEqual(Link.objects.count(), 0)

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_002_link_follow(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock(
            {
                "http://127.0.0.1/": b'Root <a href="/page1/">Link1</a>',
                "http://127.0.0.1/page1/": b'Page1 <a href="http://127.0.0.2/">Link1</a>',
                "http://127.0.0.2/": b'No 2  <a href="http://127.0.0.2/">No 2 Link2</a>',
            }
        )
        self._crawl()
        self.assertTrue(
            BrowserRequest.call_args_list
            == self.get_default_gets() + [mock.call("http://127.0.0.1/page1/", self.collection)],
            BrowserRequest.call_args_list,
        )

        self.assertEqual(Document.objects.count(), 2)
        docs = Document.objects.w_content().order_by("id")
        self.assertEqual(docs[0].url, "http://127.0.0.1/")
        self.assertEqual(docs[0].collection, self.collection)
        self.assertEqual(docs[0].content, "Root Link1")
        self.assertEqual(docs[0].crawl_recurse, 0)
        self.assertEqual(docs[1].url, "http://127.0.0.1/page1/")
        self.assertEqual(docs[1].collection, self.collection)
        self.assertEqual(docs[1].content, "Page1 Link1")
        self.assertEqual(docs[1].crawl_recurse, 0)

        self.assertEqual(Link.objects.count(), 1)
        link = Link.objects.get()
        self.assertEqual(link.doc_from, docs[0])
        self.assertEqual(link.doc_to, docs[1])
        self.assertEqual(link.text, "Link1")
        self.assertEqual(link.pos, 5)
        self.assertEqual(link.link_no, 0)

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_003_recursion_depth(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock(
            {
                "http://127.0.0.1/": b'Root <a href="/page1/">Link1</a>',
                "http://127.0.0.1/page1/": b'Page1 <a href="http://127.0.0.2/">Link1</a><a href="http://127.0.0.3/">Link3</a>',
                "http://127.0.0.2/": b'No 2  <a href="http://127.0.0.2/page1/">No 2 Link1</a><a href="http://127.0.0.3/">Link3</a>',
                "http://127.0.0.2/page1/": b'Page2 <a href="http://127.0.0.2/page2/">No 2 Link2</a>',
                "http://127.0.0.2/page2/": b"No 2 - Page2",
                "http://127.0.0.3/": b"Page3",
            }
        )
        self.collection.unlimited_regex = r"http://127\.0\.0\.1/"
        self.collection.limited_regex = r"http://127\.0\.0\.2/"
        self.collection.recursion_depth = 2
        self.collection.save()
        self._crawl()

        self.assertTrue(
            BrowserRequest.call_args_list
            == self.get_default_gets()
            + [
                mock.call("http://127.0.0.1/page1/", self.collection),
                mock.call("http://127.0.0.2/robots.txt", self.collection, check_status=True),
                mock.call("http://127.0.0.2/", self.collection),
                mock.call("http://127.0.0.2/favicon.ico", self.collection, check_status=True),
                mock.call("http://127.0.0.2/page1/", self.collection),
            ],
            BrowserRequest.call_args_list,
        )

        self.assertEqual(Document.objects.count(), 4)
        docs = Document.objects.w_content().order_by("id")
        self.assertEqual(docs[0].url, "http://127.0.0.1/")
        self.assertEqual(docs[0].collection, self.collection)
        self.assertEqual(docs[0].content, "Root Link1")
        self.assertEqual(docs[0].crawl_recurse, 0)
        self.assertEqual(docs[1].url, "http://127.0.0.1/page1/")
        self.assertEqual(docs[1].collection, self.collection)
        self.assertEqual(docs[1].content, "Page1 Link1 Link3")
        self.assertEqual(docs[1].crawl_recurse, 0)
        self.assertEqual(docs[2].url, "http://127.0.0.2/")
        self.assertEqual(docs[2].collection, self.collection)
        self.assertEqual(docs[2].content, "No 2 No 2 Link1 Link3")
        self.assertEqual(docs[2].crawl_recurse, 1)
        self.assertEqual(docs[3].url, "http://127.0.0.2/page1/")
        self.assertEqual(docs[3].collection, self.collection)
        self.assertEqual(docs[3].content, "Page2 No 2 Link2")
        self.assertEqual(docs[3].crawl_recurse, 0)

        self.assertEqual(Link.objects.count(), 3)
        links = Link.objects.order_by("id")
        self.assertEqual(links[0].doc_from, docs[0])
        self.assertEqual(links[0].doc_to, docs[1])
        self.assertEqual(links[0].text, "Link1")
        self.assertEqual(links[0].pos, 5)
        self.assertEqual(links[0].link_no, 0)
        self.assertEqual(links[1].doc_from, docs[1])
        self.assertEqual(links[1].doc_to, docs[2])
        self.assertEqual(links[1].text, "Link1")
        self.assertEqual(links[1].pos, 6)
        self.assertEqual(links[1].link_no, 0)
        self.assertEqual(links[2].doc_from, docs[2])
        self.assertEqual(links[2].doc_to, docs[3])
        self.assertEqual(links[2].text, "No 2 Link1")
        self.assertEqual(links[2].pos, 5)
        self.assertEqual(links[2].link_no, 0)

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_004_extern_links(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock(
            {
                "http://127.0.0.1/": b'Root <a href="/page1/">Link1</a>',
                "http://127.0.0.1/page1/": b"Page1",
            }
        )
        self.collection.unlimited_regex = r"http://127\.0\.0\.1/$"
        self.collection.store_extern_links = True
        self.collection.save()
        self._crawl()
        self.assertTrue(
            BrowserRequest.call_args_list == self.get_default_gets(),
            BrowserRequest.call_args_list,
        )

        self.assertEqual(Document.objects.count(), 1)
        docs = Document.objects.w_content().order_by("id")
        self.assertEqual(docs[0].url, "http://127.0.0.1/")
        self.assertEqual(docs[0].collection, self.collection)
        self.assertEqual(docs[0].content, "Root Link1")
        self.assertEqual(docs[0].crawl_recurse, 0)

        self.assertEqual(Link.objects.count(), 1)
        link = Link.objects.get()
        self.assertEqual(link.doc_from, docs[0])
        self.assertEqual(link.doc_to, None)
        self.assertEqual(link.text, "Link1")
        self.assertEqual(link.pos, 5)
        self.assertEqual(link.link_no, 0)
        self.assertEqual(link.extern_url, "http://127.0.0.1/page1/")

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("se.document.now")
    def test_005_recrawl_none(self, now, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        now.side_effect = lambda: self.fake_now
        self.collection.recrawl_freq = Collection.RECRAWL_FREQ_NONE
        self.collection.recrawl_dt_min = None
        self.collection.recrawl_dt_max = None
        self.collection.save()

        self._crawl()
        self.assertTrue(
            BrowserRequest.call_args_list == self.get_default_gets(),
            BrowserRequest.call_args_list,
        )

        self.assertEqual(Document.objects.count(), 1)
        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "Hello world")
        self.assertEqual(doc.crawl_first, self.fake_now)
        self.assertEqual(doc.crawl_last, self.fake_now)
        self.assertEqual(doc.crawl_next, None)
        self.assertEqual(doc.crawl_dt, None)

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("se.document.now")
    def test_006_recrawl_constant(self, now, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        self.collection.recrawl_freq = Collection.RECRAWL_FREQ_CONSTANT
        self.collection.recrawl_dt_min = timedelta(hours=1)
        self.collection.recrawl_dt_max = None
        self.collection.save()

        now.side_effect = lambda: self.fake_now
        self._crawl()

        self.assertEqual(Document.objects.count(), 1)
        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "Hello world")
        self.assertEqual(doc.crawl_first, self.fake_now)
        self.assertEqual(doc.crawl_last, self.fake_now)
        self.assertEqual(doc.crawl_next, self.fake_next)
        self.assertEqual(doc.crawl_dt, None)

        now.side_effect = lambda: self.fake_next
        self._crawl()

        self.assertEqual(Document.objects.count(), 1)
        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "Hello world")
        self.assertEqual(doc.crawl_first, self.fake_now)
        self.assertEqual(doc.crawl_last, self.fake_next)
        self.assertEqual(doc.crawl_next, self.fake_next + timedelta(hours=1))
        self.assertEqual(doc.crawl_dt, None)

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("se.document.now")
    def test_007_recrawl_adaptive(self, now, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        self.collection.recrawl_freq = Collection.RECRAWL_FREQ_ADAPTIVE
        self.collection.recrawl_dt_min = timedelta(hours=1)
        self.collection.recrawl_dt_max = timedelta(hours=3)
        self.collection.save()

        now.side_effect = lambda: self.fake_now
        self._crawl()

        self.assertEqual(Document.objects.count(), 1)
        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "Hello world")
        self.assertEqual(doc.crawl_first, self.fake_now)
        self.assertEqual(doc.crawl_last, self.fake_now)
        self.assertEqual(doc.crawl_next, self.fake_next)
        self.assertEqual(doc.crawl_dt, timedelta(hours=1))

        now.side_effect = lambda: self.fake_next
        self._crawl()

        self.assertEqual(Document.objects.count(), 1)
        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.content, "Hello world")
        self.assertEqual(doc.crawl_first, self.fake_now)
        self.assertEqual(doc.crawl_last, self.fake_next)
        self.assertEqual(doc.crawl_next, self.fake_next2)
        self.assertEqual(doc.crawl_dt, timedelta(hours=2))

        now.side_effect = lambda: self.fake_next2
        self._crawl()

        self.assertEqual(Document.objects.count(), 1)
        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "Hello world")
        self.assertEqual(doc.crawl_first, self.fake_now)
        self.assertEqual(doc.crawl_last, self.fake_next2)
        self.assertEqual(doc.crawl_next, self.fake_next3)
        self.assertEqual(doc.crawl_dt, timedelta(hours=3))

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_008_base_header(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock(
            {
                "http://127.0.0.1/": b"""
                <html>
                    <head><base href="/base/" /></head>
                    <body>
                        <a href="test">base test</a>
                    </body>
                </html>
                """,
                "http://127.0.0.1/base/test": b"test page",
            }
        )

        self._crawl()

        self.assertEqual(Document.objects.count(), 2)
        doc1, doc2 = Document.objects.w_content().order_by("id")
        self.assertEqual(doc1.url, "http://127.0.0.1/")
        self.assertEqual(doc1.collection, self.collection)
        self.assertEqual(doc1.content, "base test")
        self.assertEqual(doc2.url, "http://127.0.0.1/base/test")
        self.assertEqual(doc2.collection, self.collection)
        self.assertEqual(doc2.content, "test page")

    @override_settings(TEST_MODE=False)
    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_010_generic_exception_handling(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": Exception("Generic exception")})

        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "")
        self.assertIn("Generic exception", doc.error)
        self.assertIn("Traceback (most recent call last):", doc.error)

    @override_settings(TEST_MODE=False)
    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_020_skip_exception_handling(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": SkipIndexing("Skip test")})

        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "")
        self.assertIn("Skip test", doc.error)
        self.assertNotIn("Traceback (most recent call last):", doc.error)

    @override_settings(TEST_MODE=False)
    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_030_auth_failed_exception_handling(self, BrowserRequest):
        page = Page("http://127.0.0.1/", b"test", BrowserMock)
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": AuthElemFailed(page, "xxx not found")})

        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "")
        self.assertIn("xxx not found", doc.error)
        self.assertNotIn("Traceback (most recent call last):", doc.error)

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_040_extern_url_update(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock(
            {
                "http://127.0.0.1/": b'<html><body><a href="/extern.html">extern link</a></body></html>',
                "http://127.0.0.1/extern.html": b"<html><body>extern page</body></html>",
            }
        )

        self.collection.unlimited_regex = "http://127.0.0.1/$"
        self.collection.unlimited_regex_pg = "http://127.0.0.1/$"
        self.collection.store_extern_links = True
        self.collection.save()

        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "extern link")

        self.assertEqual(Link.objects.count(), 1)
        link = Link.objects.get()
        self.assertEqual(link.doc_from, doc)
        self.assertIsNone(link.doc_to)
        self.assertEqual(link.extern_url, "http://127.0.0.1/extern.html")

        self.collection.unlimited_regex = "http://127.0.0.1/.*$"
        self.collection.unlimited_regex_pg = "http://127.0.0.1/.*$"
        self.collection.save()

        self._crawl("http://127.0.0.1/extern.html")

        self.assertEqual(Document.objects.count(), 2)
        doc1 = Document.objects.w_content().order_by("id").first()
        self.assertEqual(doc1.url, "http://127.0.0.1/")
        self.assertEqual(doc1.collection, self.collection)
        self.assertEqual(doc1.content, "extern link")
        doc2 = Document.objects.w_content().order_by("id").last()
        self.assertEqual(doc2.url, "http://127.0.0.1/extern.html")
        self.assertEqual(doc2.collection, self.collection)
        self.assertEqual(doc2.content, "extern page")

        self.assertEqual(Link.objects.count(), 1)
        link = Link.objects.get()
        self.assertEqual(link.doc_from, doc1)
        self.assertEqual(link.doc_to, doc2)
        self.assertIsNone(link.extern_url)

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_050_binary_indexing(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({})
        self.collection.mimetype_regex = ".*"
        self.collection.save()
        self._crawl("http://127.0.0.1/image.png")

    MAILTO = {"http://127.0.0.1/": b'<body><a href="mailto:test@exemple.com">mail</a></body>'}

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_060_invalid_scheme(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock(self.MAILTO)
        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.content, "mail")
        self.assertEqual(Link.objects.count(), 0)

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_070_unknown_scheme_store(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock(self.MAILTO)
        self.collection.store_extern_links = True
        self.collection.save()
        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "mail")

        self.assertEqual(Link.objects.count(), 1)
        link = Link.objects.get()
        self.assertEqual(link.doc_from, doc)
        self.assertEqual(link.text, "mail")
        self.assertEqual(link.extern_url, "mailto:test@exemple.com")

    INVALID_LINK = {"http://127.0.0.1/": b'<body><a href="http://[invalid IPV6/">link</a></body>'}

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_090_invalid_link(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock(self.INVALID_LINK)
        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "link")
        self.assertEqual(Link.objects.count(), 0)

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_100_invalid_link_store(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock(self.INVALID_LINK)
        self.collection.store_extern_links = True
        self.collection.save()
        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

        doc = Document.objects.w_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "link")

        self.assertEqual(Link.objects.count(), 1)
        link = Link.objects.get()
        self.assertEqual(link.doc_from, doc)
        self.assertEqual(link.text, "link")
        self.assertEqual(link.extern_url, "http://[invalid IPV6/")

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_110_excluded_url(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock(
            {"http://127.0.0.1/": b'<html><body><a href="/page.html">link</a></body></html>'}
        )
        ExcludedUrl.objects.create(url="http://127.0.0.1/page.html")
        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_120_excluded_url_starts_with(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock(
            {
                "http://127.0.0.1/": b'<html><body><a href="/no/exclude.html">link</a><a href="/page.html">link</a></body></html>'
            }
        )
        ExcludedUrl.objects.create(url="http://127.0.0.1/no/", starting_with=True)
        self._crawl()

        self.assertEqual(Document.objects.count(), 2)
        urls = Document.objects.wo_content().values_list("url", flat=True).order_by("url")
        self.assertEqual(list(urls), ["http://127.0.0.1/", "http://127.0.0.1/page.html"])

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("se.document.now")
    def test_130_reschedule_no_change(self, now, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"<html><body>aaa 42</body></html>"})
        self.collection.recrawl_freq = Collection.RECRAWL_FREQ_ADAPTIVE
        self.collection.recrawl_dt_min = timedelta(hours=1)
        self.collection.recrawl_dt_max = timedelta(hours=3)
        self.collection.save()

        now.side_effect = lambda: self.fake_now
        self._crawl()
        doc = Document.objects.w_content().get()
        self.assertEqual(doc.content, "aaa 42")
        self.assertEqual(doc.crawl_dt, timedelta(hours=1))

        BrowserRequest.side_effect = BrowserMock(
            {"http://127.0.0.1/": b"<html><body><!-- content unchanged -->aaa 24</body></html>"}
        )
        now.side_effect = lambda: self.fake_next
        self._crawl()
        doc = Document.objects.w_content().get()
        self.assertEqual(doc.content, "aaa 24")
        self.assertEqual(doc.crawl_dt, timedelta(hours=2))

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("se.document.now")
    def test_140_reschedule_with_change(self, now, BrowserRequest):
        self.test_130_reschedule_no_change()

        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"<html><body>bbb</body></html>"})

        now.side_effect = lambda: self.fake_next2
        self._crawl()
        doc = Document.objects.w_content().get()
        self.assertEqual(doc.content, "bbb")
        self.assertEqual(doc.crawl_dt, timedelta(hours=1))

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_150_link_nested_text(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock(
            {
                "http://127.0.0.1/": b'Root <a href="/page1/"><span>Nested</span></a>',
                "http://127.0.0.1/page1/": b"Page1",
            }
        )
        self._crawl()
        self.assertTrue(
            BrowserRequest.call_args_list
            == self.get_default_gets() + [mock.call("http://127.0.0.1/page1/", self.collection)],
            BrowserRequest.call_args_list,
        )

        self.assertEqual(Document.objects.count(), 2)
        docs = Document.objects.w_content().order_by("id")
        self.assertEqual(docs[0].url, "http://127.0.0.1/")
        self.assertEqual(docs[0].collection, self.collection)
        self.assertEqual(docs[0].content, "Root Nested")
        self.assertEqual(docs[0].crawl_recurse, 0)
        self.assertEqual(docs[1].url, "http://127.0.0.1/page1/")
        self.assertEqual(docs[1].collection, self.collection)
        self.assertEqual(docs[1].content, "Page1")
        self.assertEqual(docs[1].crawl_recurse, 0)

        self.assertEqual(Link.objects.count(), 1)
        link = Link.objects.get()
        self.assertEqual(link.doc_from, docs[0])
        self.assertEqual(link.doc_to, docs[1])
        self.assertEqual(link.text, "Nested")
        self.assertEqual(link.pos, 5)
        self.assertEqual(link.link_no, 0)

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_160_hidden(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        self.collection.hide_documents = True
        self.collection.save()

        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

        doc = Document.objects.wo_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)
        self.assertTrue(doc.hidden)

    def _test_modification_base(self):
        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

        doc = Document.objects.wo_content().get()
        self.assertEqual(doc.url, "http://127.0.0.1/")
        self.assertEqual(doc.collection, self.collection)

        # Mark the document as modified in the past
        doc.modified_date = self.fake_yesterday

        # Force crawl_next to make it crawled again when calling self._crawl()
        doc.crawl_next = self.fake_yesterday
        doc.save()

    @mock.patch("se.browser_request.BrowserRequest.get")
    def test_180_modification_no_change(self, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world 1"})
        self.collection.hash_mode = Collection.HASH_RAW
        self.collection.save()
        self._test_modification_base()

        self.assertEqual(Document.objects.count(), 1)
        doc = Document.objects.w_content().get()
        self.assertEqual(doc.content, "Hello world 1")
        self.assertEqual(doc.content_hash, md5(b"Hello world 1").hexdigest())
        self.assertEqual(doc.modified_date, self.fake_yesterday)

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("se.document.now")
    def test_190_modification_change(self, now, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world 1"})
        now.side_effect = lambda: self.fake_now
        self.collection.hash_mode = Collection.HASH_RAW
        self.collection.save()
        self._test_modification_base()

        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world 2"})
        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

        doc = Document.objects.w_content().get()
        self.assertEqual(doc.content, "Hello world 2")
        self.assertEqual(doc.content_hash, md5(b"Hello world 2").hexdigest())
        self.assertEqual(doc.modified_date, self.fake_now)

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("se.document.now")
    def test_200_modification_no_change_number_normalize(self, now, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world 1"})
        now.side_effect = lambda: self.fake_now
        self.collection.hash_mode = Collection.HASH_NO_NUMBERS
        self.collection.save()
        self._test_modification_base()

        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world 2"})
        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

        doc = Document.objects.w_content().get()
        self.assertEqual(doc.content, "Hello world 2")
        self.assertEqual(doc.content_hash, md5(b"Hello world 0").hexdigest())
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.modified_date, self.fake_yesterday)

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("se.document.now")
    def test_210_modification_change_number_normalize(self, now, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        now.side_effect = lambda: self.fake_now
        self._test_modification_base()

        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world change"})
        self._crawl()
        self.assertEqual(Document.objects.count(), 1)

        doc = Document.objects.w_content().get()
        self.assertEqual(doc.collection, self.collection)
        self.assertEqual(doc.content, "Hello world change")
        self.assertEqual(doc.content_hash, md5(b"Hello world change").hexdigest())
        self.assertEqual(doc.modified_date, self.fake_now)

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("feedparser.parse", wraps=parse)
    def test_220_recrawl_cond_on_change__no_change(self, feedparser_parse, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        self._test_modification_base()
        self.assertEqual(Document.objects.count(), 1)
        self.assertEqual(feedparser_parse.call_count, 1)

        self._crawl()
        self.assertEqual(feedparser_parse.call_count, 1)

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("feedparser.parse", wraps=parse)
    def test_230_recrawl_cond_on_change__change(self, feedparser_parse, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        self._test_modification_base()
        self.assertEqual(Document.objects.count(), 1)
        self.assertEqual(feedparser_parse.call_count, 1)

        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world change"})
        self._crawl()
        self.assertEqual(feedparser_parse.call_count, 2)

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("feedparser.parse", wraps=parse)
    def test_240_recrawl_cond_always(self, feedparser_parse, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        self.collection.recrawl_condition = Collection.RECRAWL_COND_ALWAYS
        self.collection.save()

        self._test_modification_base()
        self.assertEqual(Document.objects.count(), 1)
        self.assertEqual(feedparser_parse.call_count, 1)

        self._crawl()
        self.assertEqual(feedparser_parse.call_count, 2)

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("feedparser.parse", wraps=parse)
    def test_250_recrawl_cond_manual__change(self, feedparser_parse, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        self.collection.recrawl_condition = Collection.RECRAWL_COND_MANUAL
        self.collection.save()

        self._test_modification_base()
        self.assertEqual(Document.objects.count(), 1)
        self.assertEqual(feedparser_parse.call_count, 1)

        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world change"})
        self._crawl()
        self.assertEqual(feedparser_parse.call_count, 2)

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("feedparser.parse", wraps=parse)
    def test_260_recrawl_cond_manual__no_change(self, feedparser_parse, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        self.collection.recrawl_condition = Collection.RECRAWL_COND_MANUAL
        self.collection.save()
        self._test_modification_base()
        self.assertEqual(Document.objects.count(), 1)
        self.assertEqual(feedparser_parse.call_count, 1)

        self._crawl()
        self.assertEqual(feedparser_parse.call_count, 1)

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("feedparser.parse", wraps=parse)
    def test_270_recrawl_cond_manual__manual(self, feedparser_parse, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        self.collection.recrawl_condition = Collection.RECRAWL_COND_MANUAL
        self.collection.save()
        self._test_modification_base()
        self.assertEqual(Document.objects.count(), 1)
        self.assertEqual(feedparser_parse.call_count, 1)

        Document.objects.update(manual_crawl=True)
        self._crawl()

        self.assertEqual(feedparser_parse.call_count, 2)
        self.assertFalse(Document.objects.wo_content().get().manual_crawl)

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("feedparser.parse", wraps=parse)
    def test_280_recrawl_keeps_tags(self, feedparser_parse, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        self.collection.tags.create(name="tag1")

        self.assertEqual(Document.objects.count(), 0)
        self._crawl()
        self.assertEqual(Document.objects.count(), 1)
        doc = Document.objects.w_content().get()
        self.assertEqual(list(doc.tags.values_list("name", flat=True)), ["tag1"])

        doc.tags.create(name="tag2")
        self.assertEqual(list(doc.tags.values_list("name", flat=True)), ["tag1", "tag2"])

        self._crawl()
        doc = Document.objects.w_content().get()
        self.assertEqual(list(doc.tags.values_list("name", flat=True)), ["tag1", "tag2"])

    @mock.patch("se.browser_request.BrowserRequest.get")
    @mock.patch("se.document.now")
    def test_290_recrawl_new_pages_first(self, now, BrowserRequest):
        BrowserRequest.side_effect = BrowserMock({"http://127.0.0.1/": b"Hello world"})
        now.side_effect = lambda: self.fake_now

        self.collection.recrawl_freq = Collection.RECRAWL_FREQ_NONE
        self.collection.save()

        already_crawled = Document.objects.create(
            collection=self.collection,
            url="http://127.0.0.1/page.html",
            content="done",
            crawl_last=self.fake_yesterday,
            crawl_next=self.fake_yesterday,
        )

        WorkerStats.get_worker(0)
        Document.queue("http://127.0.0.1/", self.collection, None)
        Document.crawl(0)

        self.assertEqual(Document.objects.count(), 2)

        doc = Document.objects.w_content().get(url="http://127.0.0.1/")
        self.assertEqual(doc.crawl_last, self.fake_now)

        already_crawled.refresh_from_db()
        self.assertEqual(already_crawled.crawl_last, self.fake_yesterday)
        self.assertEqual(already_crawled.crawl_next, self.fake_yesterday)

    def test_300_queue_to_any_collection_enabled(self):
        # Create two collections with different regex patterns (avoiding conflicts with setUp)
        coll1 = Collection.objects.create(
            name="Collection 1", unlimited_regex="http://example1.com/", queue_to_any_collection=True
        )
        coll2 = Collection.objects.create(name="Collection 2", unlimited_regex="http://example2.com/")

        # Create a parent document in collection 1
        parent_doc = Document.queue("http://example1.com/parent", coll1, None)

        # Try to queue a URL that doesn't match coll1 but matches coll2
        queued_doc = Document.queue("http://example2.com/child", coll1, parent_doc)

        # Should be queued in collection 2, not collection 1
        self.assertIsNotNone(queued_doc)
        self.assertEqual(queued_doc.collection, coll2)
        self.assertEqual(queued_doc.url, "http://example2.com/child")

    def test_310_cross_collection_crawl_disabled(self):
        # Create two collections with different regex patterns
        coll1 = Collection.objects.create(
            name="Collection 1", unlimited_regex="http://site1.com/", queue_to_any_collection=False
        )
        Collection.objects.create(name="Collection 2", unlimited_regex="http://site2.com/")

        # Create a parent document in collection 1
        parent_doc = Document.queue("http://site1.com/parent", coll1, None)

        # Try to queue a URL that doesn't match coll1 but matches coll2
        queued_doc = Document.queue("http://site2.com/child", coll1, parent_doc)

        # Should return None (not queued) when cross-collection crawl is disabled
        self.assertIsNone(queued_doc)

    def test_320_cross_collection_crawl_same_collection(self):
        # Create a collection
        coll1 = Collection.objects.create(
            name="Collection 1", unlimited_regex="http://site1.com/", queue_to_any_collection=True
        )

        # Create a parent document
        parent_doc = Document.queue("http://site1.com/parent", coll1, None)

        # Try to queue a URL that matches the same collection
        queued_doc = Document.queue("http://site1.com/child", coll1, parent_doc)

        # Should be queued normally in the same collection
        self.assertIsNotNone(queued_doc)
        self.assertEqual(queued_doc.collection, coll1)

    def test_330_cross_collection_crawl_no_match(self):
        # Create one collection
        coll1 = Collection.objects.create(
            name="Collection 1", unlimited_regex="http://site1.com/", queue_to_any_collection=True
        )

        # Create a parent document
        parent_doc = Document.queue("http://site1.com/parent", coll1, None)

        # Try to queue a URL that doesn't match any collection
        queued_doc = Document.queue("http://nomatch.com/child", coll1, parent_doc)

        # Should return None (not queued)
        self.assertIsNone(queued_doc)

    def test_340_cross_collection_crawl_selective(self):
        # Create three collections with different regex patterns
        coll1 = Collection.objects.create(name="Collection 1", unlimited_regex="http://site1.com/")
        coll2 = Collection.objects.create(name="Collection 2", unlimited_regex="http://site2.com/")
        Collection.objects.create(name="Collection 3", unlimited_regex="http://site3.com/")

        # Configure coll1 to only crawl cross-collection to coll2 (not coll3)
        coll1.queue_to_collections.add(coll2)

        # Create a parent document in collection 1
        parent_doc = Document.queue("http://site1.com/parent", coll1, None)

        # Try to queue a URL that matches coll2 (should work)
        queued_doc2 = Document.queue("http://site2.com/child", coll1, parent_doc)
        self.assertIsNotNone(queued_doc2)
        self.assertEqual(queued_doc2.collection, coll2)

        # Try to queue a URL that matches coll3 (should NOT work since coll3 not in selective list)
        queued_doc3 = Document.queue("http://site3.com/child", coll1, parent_doc)
        self.assertIsNone(queued_doc3)

    def test_350_queue_to_any_collection_priority_over_selective(self):
        # Create three collections
        coll1 = Collection.objects.create(
            name="Collection 1",
            unlimited_regex="http://site1.com/",
            queue_to_any_collection=True,  # This takes priority over selective configuration
        )
        coll2 = Collection.objects.create(name="Collection 2", unlimited_regex="http://site2.com/")
        coll3 = Collection.objects.create(name="Collection 3", unlimited_regex="http://site3.com/")

        # Configure selective crawl (should be ignored when queue_to_any_collection=True)
        coll1.queue_to_collections.add(coll2)

        # Create a parent document in collection 1
        parent_doc = Document.queue("http://site1.com/parent", coll1, None)

        # Try to queue a URL that matches coll2 (should work - matches and all mode is on)
        queued_doc2 = Document.queue("http://site2.com/child", coll1, parent_doc)
        self.assertIsNotNone(queued_doc2)
        self.assertEqual(queued_doc2.collection, coll2)

        # Try to queue a URL that matches coll3 (should ALSO work since all mode takes priority)
        queued_doc3 = Document.queue("http://site3.com/child", coll1, parent_doc)
        self.assertIsNotNone(queued_doc3)
        self.assertEqual(queued_doc3.collection, coll3)
